#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .optimizer import Optimizer, CandidateState, OptimizationResults 
import numpy as np

class NelderMead(Optimizer):
    """Squirrel Search Algorithm class"""
    
    def __init__(self):
        """Initialization"""
        Optimizer.__init__(self)
        
        self.X = None
        self.X0 = None
        self.method = 'Vanilla'
        self.params = {}
        self.params['init_step'] = 0.4
        self.params['alpha'] = 1.0
        self.params['gamma'] = 2.0
        self.params['rho'] = 0.5
        self.params['sigma'] = 0.5
        self.iterations = 100
    
    
    def init_params(self):

        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []

        if self.method == 'Vanilla':
            mandatory_params = ''.split()
            optional_params = 'init_step alpha gamma rho sigma'.split()

        for param in mandatory_params:
            # if param not in defined_params:
            #    print('Error: Missing parameter (%s)' % param)
            assert param in defined_params, f'Error: Missing parameter {param}'

        for param in defined_params:
            if param not in mandatory_params and param not in optional_params:
                print(f'Warning: Excessive parameter {param}')
                
                
    def _init_optimizer(self):
        
        super(NelderMead, self)._init_optimizer()
        
        # Bounds for position and velocity
        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)

        # Generate set of points
        self.cS = np.array([CandidateState(self) for _ in range(self.dimensions + 1)], \
                            dtype=CandidateState)
            
        # Generate initial positions
        #self.cS[0].X = 0.5 * (self.lb + self.ub)
        self.cS[0].X = np.random.uniform(self.lb, self.ub)
        if self.X0 is not None:
                if np.shape(self.X0)[0] > 0:
                    self.cS[0].X = self.X0[0]
        self.cS[0].evaluate()
        print(self.cS[0].X)
         
        for p in range(1, self.dimensions + 1):
            
            # Random position
            dx = np.zeros([self.dimensions])
            dx[p - 1] = self.params['init_step']
            self.cS[p].X = self.cS[p].X + dx * (self.ub - self.lb)
            
            # Using specified particles initial positions
            if self.X0 is not None:
                if p < np.shape(self.X0)[0]:
                    self.cS[p].X = self.X0[p]
                                          
            # Evaluate
            self.cS[p].evaluate()

    def run(self, seed=None):
        super(LegacyPSO, self).run(seed=seed)

        self.init_params()
        self._init_optimizer()
        
        
        for i in range(self.iterations):
            self.cS = np.sort(self.cS)
            print(i, np.min(self.cS).f)
            #print(i, self.cS[0].f, self.cS[-1].f)


            # Center
            X0= np.zeros(self.dimensions)
            for p in range(self.dimensions):
                X0 += self.cS[p].X
            X0 /= self.dimensions
            
            dX = X0 - self.cS[-1].X
            
            # Reflection
            Xr = X0 + self.params['alpha'] * dX
            cR = CandidateState(self)
            cR.X = Xr
            cR.evaluate()
                       
            if self.cS[0] <= cR <= self.cS[2]:
                self.cS[-1] = cR.copy()
                #print('Rf')
                continue
                
            
            # Expansion
            if cR < self.cS[0]:
                Xe = X0 + self.params['gamma'] * dX
                cE = CandidateState(self)
                cE.X = Xe
                cE.evaluate()
            
                if cE < cR:
                    self.cS[-1] = cE.copy()
                    #print('Ex')
                    continue
                else:
                    self.cS[-1] = cR.copy()
                    #print('Rf')
                    continue


            # Contraction
            if cR < self.cS[-1]:
                
                Xc = X0 + self.params['rho'] * dX
                cC = CandidateState(self)
                cC.X = Xc
                cC.evaluate()
                
                if cC < self.cS[-1]:
                    self.cS[-1] = cC.copy()
                    #print('Ct')
                    continue
                
            else:
                
                Xc = X0 - self.params['rho'] * dX
                cC = CandidateState(self)
                cC.X = Xc
                cC.evaluate()
                
                if cC < self.cS[-1]:
                    self.cS[-1] = cC.copy()
                    #print('Ct')
                    continue

            
            # Reduction
            for p in range(1, self.dimensions + 1):
                self.cS[p].X = self.cS[0].X + self.params['sigma'] * (self.cS[p].X - self.cS[0].X)
                self.cS[p].evaluate()
            #print('Rd')

        return np.min(self.cS)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 27 12:19:49 2018

@author: stefan
"""

import numpy as np
import os


class CEC2014():
    """CEC'14 Test Functions"""

    def __init__(self, d):
        """Initialize and read all data"""

        data_dir = os.path.dirname(os.path.abspath(__file__))
        # print('Reading data from', data_dir)
        self.M, self.O, self.S = [], [], []
        for f in range(30):
            self.M.append(np.loadtxt(
                '%s/cec2014_data/M_%d_D%d.txt' % (data_dir, f + 1, d)))
            self.O.append(np.loadtxt(
                '%s/cec2014_data/shift_data_%d.txt' % (data_dir, f + 1))[:d])
            self.S.append(np.loadtxt('%s/cec2014_data/shuffle_data_%d_D%d.txt' %
                                     (data_dir, f + 1, d), dtype='int') - 1)
            """
            print('f%02d %dD' % (f + 1, d))
            print('M: ', self.M[f].shape)
            print('O: ', self.O[f].shape)
            print('S: ', self.S[f].shape)
            """
        self.functions = [self.F1, self.F2, self.F3, self.F4, self.F5, self.F6, 
                          self.F7, self.F8, self.F9, self.F10, self.F11, 
                          self.F12, self.F13, self.F14, self.F15, self.F16,
                          self.F17, self.F18, self.F19, self.F20, self.F21, 
                          self.F22, self.F23, self.F24, self.F25, self.F26,
                          self.F27, self.F28, self.F29, self.F30]

    def f1(self, x):
        """High Conditioned Elliptic Function"""
        d = np.size(x)
        return np.sum(np.power(1e6, np.arange(d) / (d - 1)) * np.power(x, 2))

    def f2(self, x):
        """Bent Cigar Function"""
        return x[0]**2 + 1e6 * np.sum(x[1:]**2)

    def f3(self, x):
        """Discus Function"""
        return 1e6 * x[0]**2 + np.sum(x[1:]**2)

    def f4(self, x):
        """Rosenbrock’s Function"""
        return np.sum(100.0 * (x[:-1]**2 - x[1:])**2 + (x[:-1] - 1.0)**2)

    def f5(self, x):
        """Ackley’s Function"""
        d = np.size(x)
        return -20.0 * np.exp(-0.2 * np.sqrt(np.sum(x**2) / d)) - \
            np.exp(1.0 / d * np.sum(np.cos(2 * np.pi * x))) + 20 + np.e

    def f6(self, x):
        """Weierstrass Function"""
        a, b, kmax = 0.5, 3.0, 20
        d = np.size(x)
        s1, s2 = 0.0, 0.0
        for k in range(kmax + 1):
            s1 += np.power(a, k) * np.cos(2 * np.pi *
                                          np.power(b, k) * (x + 0.5))
            s2 += np.power(a, k) * np.cos(2 * np.pi * np.power(b, k) * 0.5)
        return np.sum(s1) - d * s2

    def f7(self, x):
        """Griewank’s Function"""
        d = np.size(x)
        return np.sum(x**2 / 4000.0) - \
            np.prod(np.cos(x / np.sqrt(np.arange(d) + 1))) + 1.0

    def f8(self, x):
        """Rastrigin's Function"""
        return np.sum(x**2 - 10.0 * np.cos(2 * np.pi * x) + 10)

    def f9(self, x):
        """Modified Schwefel’s Function"""
        d = np.size(x)
        z = x + 4.209687462275036e+002
        g = np.empty_like(x)
        g[np.abs(z) <= 500.0] = z[np.abs(z) <= 500.0] * \
            np.sin(np.sqrt(np.abs(z[np.abs(z) <= 500.0])))
        m500 = 500.0 - np.mod(z[z > 500.0], 500.0)
        g[z > 500.0] = m500 * \
            np.sin(np.sqrt(np.abs(m500))) - \
            (z[z > 500.0] - 500) ** 2 / (10000 * d)
        ma500 = np.mod(np.abs(z[z < -500.0]), 500.0) - 500.0
        g[z < -500.0] = ma500 * \
            np.sin(np.sqrt(np.abs(ma500))) - \
            (z[z < -500.0] + 500.0)**2 / (10000 * d)
        return 418.9829 * d - np.sum(g)

    def f10(self, x):
        """Katsuura Function"""
        d = np.size(x)
        s = np.zeros_like(x)
        for i in range(d):
            for j in range(1, 33):
                s[i] += np.abs(2**j * x[i] - np.round(2**j * x[i])) / 2**j
        return (10 / d**2) * np.prod((1 + np.arange(1, d + 1) * s)
                                     ** (10 / d**1.2)) - (10 / d**2)

    def f11(self, x):
        """HappyCat Function"""
        d = np.size(x)
        x = x - 1  # undocumented shift to origin
        return np.abs(np.sum(x**2) - d)**(1 / 4) + \
            (0.5 * np.sum(x**2) + np.sum(x)) / d + 0.5

    def f12(self, x):
        """HGBat Function"""
        d = np.size(x)
        x = x - 1  # undocumented shift to origin
        return np.abs(np.sum(x**2)**2 - np.sum(x)**2)**(1 / 2) + \
            (0.5 * np.sum(x**2) + np.sum(x)) / d + 0.5

    def f13(self, x):
        """Expanded Griewank’s plus Rosenbrock’s Function"""
        s = 0.0
        for pair in np.array([x, np.roll(x, -1)]).T:
            s += self.f7(self.f4(pair))
        return s

    def f14(self, x):
        """Expanded Scaffer’s F6 Function"""
        # TODO: Replace lambda with a function (pep8 E731)
        def g(X, Y): return 0.5 + (np.sin((X**2 + Y**2)**0.5)
                                   ** 2 - 0.5) / (1 + 0.001 * (X**2 + Y**2))**2
        """
        # Possible speedup - needs to be tested
        def gg(x, y):
            xy2 = x**2 + y**2
            return 0.5 + (np.sin(np.sqrt(xy2))**2-0.5) / (1 + 0.001 * xy2)**2
        """
        return np.sum(g(x, np.roll(x, -1)))

    def F1(self, x, M=None, O=None):
        """Rotated High Conditioned Elliptic Function"""
        if M is None:
            M = self.M[0]

        if O is None:
            O = self.O[0]
        return self.f1(np.dot(M, x - O))

    def F2(self, x, M=None, O=None):
        """Rotated Bent Cigar Function"""
        if M is None:
            M = self.M[1]

        if O is None:
            O = self.O[1]

        return self.f2(np.dot(M, x - O))

    def F3(self, x, M=None, O=None):
        """Rotated Discus Function"""
        if M is None: M = self.M[2]
        if O is None: O = self.O[2]
        return self.f3(np.dot(M, x - O))

    def F4(self, x, M=None, O=None):
        """Shifted and Rotated Rosenbrock’s Function"""
        if M is None:
            M = self.M[3]

        if O is None:
            O = self.O[3]

        return self.f4(np.dot(M, 2.048 * (x - O) / 100) + 1)

    def F5(self, x, M=None, O=None):
        """Shifted and Rotated Ackley’s Function"""
        if M is None:
            M = self.M[4]

        if O is None:
            O = self.O[4]

        return self.f5(np.dot(M, x - O))

    def F6(self, x, M=None, O=None):
        """Shifted and Rotated Weierstrass Function"""
        if M is None:
            M = self.M[5]

        if O is None:
            O = self.O[5]

        return self.f6(np.dot(M, 0.005 * (x - O)))

    def F7(self, x, M=None, O=None):
        """Shifted and Rotated Griewank’s Function"""
        if M is None:
            M = self.M[6]

        if O is None:
            O = self.O[6]

        return self.f7(np.dot(M, 6.0 * (x - O)))

    def F8(self, x, M=None, O=None):
        """Shifted Rastrigin’s Function"""
        if O is None:
            O = self.O[7]
        return self.f8(0.0512 * (x - O))

    def F9(self, x, M=None, O=None):
        """Shifted and Rotated Rastrigin’s Function"""
        if M is None:
            M = self.M[8]

        if O is None:
            O = self.O[8]
        return self.f8(np.dot(M, 0.0512 * (x - O)))

    def F10(self, x, M=None, O=None):
        """Shifted Schwefel’s Function"""
        if O is None:
            O = self.O[9]
        return self.f9(10.0 * (x - O))

    def F11(self, x, M=None, O=None):
        """Shifted and Rotated Schwefel’s Function"""
        if M is None:
            M = self.M[10]

        if O is None:
            O = self.O[10]
        return self.f9(np.dot(M, 10.0 * (x - O)))

    def F12(self, x, M=None, O=None):
        """Shifted and Rotated Katsuura Function"""
        if M is None:
            M = self.M[11]

        if O is None:
            O = self.O[11]
        return self.f10(np.dot(M, 0.05 * (x - O)))

    def F13(self, x, M=None, O=None):
        """Shifted and Rotated HappyCat Function"""
        if M is None:
            M = self.M[12]

        if O is None:
            O = self.O[12]
        return self.f11(np.dot(M, 0.05 * (x - O)))

    def F14(self, x, M=None, O=None):
        """Shifted and Rotated HGBat Function"""
        if M is None:
            M = self.M[13]

        if O is None:
            O = self.O[13]
        return self.f12(np.dot(M, 0.05 * (x - O)))

    def F15(self, x, M=None, O=None):
        """Shifted and Rotated Expanded Griewank’s plus Rosenbrock’s Function"""
        if M is None:
            M = self.M[14]

        if O is None:
            O = self.O[14]
        return self.f13(np.dot(M, 0.05 * (x - O)) + 1)

    def F16(self, x, M=None, O=None):
        """Shifted and Rotated Expanded Scaffer’s F6 Function"""
        if M is None:
            M = self.M[15]

        if O is None:
            O = self.O[15]
        # return self.f14(np.dot(M, (x - O)) + 1)  # shifting by -1 is documented but not implemented in C++ code
        return self.f14(np.dot(M, x - O))

    def hybrid_function(self, caller_function, S, p, g, x):
        d = np.size(x)
        n = np.cumsum(np.array(p) * d) - np.array(p) * d
        n = np.append(n, d).astype('int')
        s = 0.0
        if caller_function in range(17, 23):
            x_sr = np.dot(self.M[caller_function - 1],
                          x - self.O[caller_function - 1])
            S = self.S[caller_function - 1]
        else:   # called by composition_function, no need to shift/rotate
            x_sr = x
        for n_start, n_stop, gg in zip(n[:-1], n[1:], g):
            indices = S[n_start:n_stop]
            s += gg(x_sr[indices], M=1.0, O=0.0)
        return s

    def F17(self, x, M=None, O=None, upstream_caller_function=None, S=None):
        """Hybrid Function 1"""
        if M is not None and O is not None:
            x = np.dot(M, x - O)
        function_nr = 17
        if upstream_caller_function is not None:  # called by composition_function
            function_nr = upstream_caller_function
        return self.hybrid_function(function_nr, S,
                                    [0.3, 0.3, 0.4],
                                    [self.F10, self.F8, self.F1], x)

    def F18(self, x, M=None, O=None, upstream_caller_function=None, S=None):
        """Hybrid Function 2"""
        if M is not None and O is not None:
            x = np.dot(M, x - O)
        function_nr = 18
        if upstream_caller_function is not None:
            function_nr = upstream_caller_function
        return self.hybrid_function(function_nr, S,
                                    [0.3, 0.3, 0.4],
                                    [self.F2, self.F14, self.F8], x)

    def F19(self, x, M=None, O=None, upstream_caller_function=None, S=None):
        """Hybrid Function 3"""
        if M is not None and O is not None:
            x = np.dot(M, x - O)
        function_nr = 19
        if upstream_caller_function is not None:
            function_nr = upstream_caller_function
        return self.hybrid_function(function_nr, S,
                                    [0.2, 0.2, 0.3, 0.3],
                                    [self.F7, self.F6, self.F4, self.F16], x)

    def F20(self, x, M=None, O=None, upstream_caller_function=None, S=None):
        """Hybrid Function 4"""
        if M is not None and O is not None:
            x = np.dot(M, x - O)
        function_nr = 20
        if upstream_caller_function is not None:
            function_nr = upstream_caller_function
        return self.hybrid_function(function_nr, S,
                                    [0.2, 0.2, 0.3, 0.3],
                                    [self.F14, self.F3, self.F15, self.F8], x)

    def F21(self, x, M=None, O=None, upstream_caller_function=None, S=None):
        """Hybrid Function 5"""
        if M is not None and O is not None:
            x = np.dot(M, x - O)
        function_nr = 21
        if upstream_caller_function is not None:
            function_nr = upstream_caller_function
        return self.hybrid_function(function_nr, S,
                                    [0.1, 0.2, 0.2, 0.2, 0.3],
                                    [self.F16, self.F14, self.F4, self.F10, self.F1], x)

    def F22(self, x, M=None, O=None, upstream_caller_function=None, S=None):
        """Hybrid Function 6"""
        if M is not None and O is not None:
            x = np.dot(M, x - O)
        function_nr = 22
        if upstream_caller_function is not None:
            function_nr = upstream_caller_function
        return self.hybrid_function(function_nr, S,
                                    [0.1, 0.2, 0.2, 0.2, 0.3],
                                    [self.F12, self.F13, self.F15, self.F11, self.F5], x)

    def composition_function(self, caller_function, sigma, lmbd, bias, g, x):
        d = np.size(x)
        w = np.empty(len(g))
        for i in range(len(g)):
            O = self.O[caller_function - 1][i, :d]
            sumsqrd = np.sum((x - O)**2)
            w[i] = (1 / np.sqrt(sumsqrd)) * \
                np.exp(-sumsqrd / (2 * d * sigma[i]**2))
        omega = w / np.sum(w)
        s = 0.0
        for i in range(len(g)):
            M = self.M[caller_function - 1][i * d:(i + 1) * d, :]
            O = self.O[caller_function - 1][i, :d]
            # calling a hybrid function asks for shuffling per caller_function
            if g[i] in [self.F17, self.F18, self.F19,
                        self.F20, self.F21, self.F22]:
                S = self.S[caller_function - 1][d * i:d * (i + 1)]
                s += omega[i] * (lmbd[i] * g[i](x, M, O,
                                                caller_function, S) + bias[i])
            else:
                s += omega[i] * (lmbd[i] * g[i](x, M, O) + bias[i])

        return s

    def F23(self, x):
        """Composition Function 1"""
        return self.composition_function(23,
                                         [10, 20, 30, 40, 50],
                                         [1, 1e-6, 1e-26, 1e-6, 1e-6],
                                         [0, 100, 200, 300, 400],
                                         [self.F4, self.F1, self.F2, self.F3,
                                          lambda x, M, O: self.F1(x, 1, O)], x)

    def F24(self, x):
        """Composition Function 2"""
        return self.composition_function(24,
                                         [20, 20, 20],
                                         [1, 1, 1],
                                         [0, 100, 200],
                                         [lambda x, M, O: self.F10(x, 1, O),
                                          self.F9, self.F14], x)

    def F25(self, x):
        """Composition Function 3"""
        return self.composition_function(25,
                                         [10, 30, 50],
                                         [0.25, 1, 1e-7],
                                         [0, 100, 200],
                                         [self.F11, self.F9, self.F1], x)

    def F26(self, x):
        """Composition Function 4"""
        return self.composition_function(26,
                                         [10, 10, 10, 10, 10],
                                         [0.25, 1, 1e-7, 2.5, 10],
                                         [0, 100, 200, 300, 400],
                                         [self.F11, self.F13, self.F1, self.F6, self.F7], x)

    def F27(self, x):
        """Composition Function 5"""
        return self.composition_function(27,
                                         [10, 10, 10, 20, 20],
                                         [10, 10, 2.5, 25, 1e-6],
                                         [0, 100, 200, 300, 400],
                                         [self.F14, self.F9, self.F11, self.F6, self.F1], x)

    def F28(self, x):
        """Composition Function 6"""
        return self.composition_function(28,
                                         [10, 20, 30, 40, 50],
                                         [2.5, 10, 2.5, 5e-4, 1e-6],
                                         [0, 100, 200, 300, 400],
                                         [self.F15, self.F13, self.F11, self.F16, self.F1], x)

    def F29(self, x):
        """Composition Function 7"""
        return self.composition_function(29,
                                         [10, 30, 50],
                                         [1, 1, 1],
                                         [0, 100, 200],
                                         [self.F17, self.F18, self.F19], x)

    def F30(self, x):
        """Composition Function 8"""
        return self.composition_function(30,
                                         [10, 30, 50],
                                         [1, 1, 1],
                                         [0, 100, 200],
                                         [self.F20, self.F21, self.F22], x)


if __name__ == '__main__':

    x = np.random.uniform(-100, 100, 10)
    cec = CEC2014(10)
    test_functions = cec.functions
#    test_functions = [cec.f1, cec.f2, cec.f3, cec.f4, cec.f5, cec.f6, cec.f7,
#                      cec.f8, cec.f9, cec.f10, cec.f11, cec.f12, cec.f13, cec.f14,
#                      cec.F1, cec.F2, cec.F3, cec.F4, cec.F5, cec.F6, cec.F7,
#                      cec.F8, cec.F9, cec.F10, cec.F11, cec.F12, cec.F13,
#                      cec.F14, cec.F15, cec.F16, cec.F17, cec.F18, cec.F19,
#                      cec.F20, cec.F21, cec.F22, cec.F23, cec.F24, cec.F25,
#                      cec.F26, cec.F27, cec.F28, cec.F29, cec.F30]

    for f in test_functions:
        print('%3s %s\n %20f' % (f.__name__, f.__doc__, f(x)))

    # import time
    # start = time.time()
    # for r in range(1000*10*10):
    #     # print(r)
    #     cec.F7(x)
    # end = time.time()
    # print(end - start)



# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import time
import multiprocessing
import string

class Optimizer:
    """Base class for all optimization methods."""

    def __init__(self):
        """Initialization"""

        self.dimensions = None
        self.evaluation_function = None
        self.number_of_processes = 1
        
        self.objectives = 1
        self.objective_weights = None
        self.objective_labels = None
        self.constraints = 0
        self.constraint_labels = None
        
        self.iterations = None
        self.target_fitness = -np.inf
        
        self.lb = None
        self.ub = None
        self.verbose = None
        self.report = {}

        self.forward_unique_str = False
        self._unique_str_list = []
        
    def _init_optimizer(self):
        
        assert not isinstance(self.objectives, np.unsignedinteger), \
            "optimizer.objectives should be positive integer"
        assert self.objectives > 0, "optimizer.objectives should be positive integer"
        
        assert not isinstance(self.constraints, np.unsignedinteger), \
            "optimizer.constraints should be unsigned integer"
        
        if self.objective_weights is None:
            self.objective_weights = np.ones(self.objectives) / self.objectives   
        else:
            assert len(self.objective_weights) == self.objectives, \
                "optimizer.objective_weights list should contain number of elements equal to optimizer.objectives"
        
        if self.objective_labels is None:
            self.objective_labels = [f'o_{o}' for o in range(self.objectives)]   
        else:
            assert len(self.objective_labels) == self.objectives, \
                "optimizer.objective_labels list should contain number of strings equal to optimizer.objectives"
            
        if self.constraint_labels is None:
            self.constraint_labels = [f'c_{c}' for c in range(self.constraints)]   
        else:
            assert len(self.constraint_labels) == self.constraints, \
                "optimizer.constraint_labels list should contain number of strings equal to optimizer.constraints"
            
        self.ET = None

        assert isinstance(self.number_of_processes, int) or self.number_of_processes == 'maximum', \
            "optimizer.number_of_processes should be integer or \'maximum\'"
        if self.number_of_processes == 'maximum':
            self.number_of_processes = multiprocessing.cpu_count()
        if self.number_of_processes > 1:
            self.log(f'Preparing a multiprocess pool for {self.number_of_processes} processes.')
            self.pool = multiprocessing.Pool(self.number_of_processes)


    def _check_params(self, mandatory_params, optional_params, defined_params):
        for param in mandatory_params:
            assert param in defined_params, f'Error: Missing parameter {param}'

        for param in defined_params:
            if param not in mandatory_params and param not in optional_params:
                print(f'Warning: Excessive parameter {param}')

    def title(self, msg, line_length=60):
        t = '-' * line_length
        t += '\n' + msg.center(line_length) + '\n'
        t += '-' * line_length
        print(t)

    def log(self, msg, indent=0):
        print(' ' * indent + msg)
        
    def tic(self):
        self._time1 = time.time()

    def toc(self, msg='', silent=False):

        s = time.time() - self._time1
        
        if not silent:
            if msg == '':
                msg = 'Elapsed time:'

            if s < 1:
                dur = f'{s*1e3:.3f} milliseconds.'
            else:
                dur = f'{s:.3f} seconds.'
            self.log(f'{msg} {dur}')
            
        return s


    def _gen_unique_str(self):
        while True:
            _s = ''.join(np.random.choice([c for c in string.ascii_lowercase]) for i in range(16))
            if _s not in self._unique_str_list:
                self._unique_str_list.append(_s)
                return _s


    def run(self, seed=None):
        if None:
            np.random.seed() # Possibe fix for some seed issues in parallel execution
        else:
            np.random.seed(seed)



    def __getstate__(self):
        self_dict = self.__dict__.copy()
        del self_dict['pool']
        return self_dict

    def __setstate__(self, state):
        self.__dict__.update(state)

    def _multiprocess_evaluate(self, candidates):

        if self.forward_unique_str:
            result = self.pool.starmap(self.evaluation_function, [tuple([c.X, self._gen_unique_str()]) for c in candidates])
        else:
            result = self.pool.map(self.evaluation_function, [c.X for c in candidates])

        if self.objectives == 1 and self.constraints == 0:
            # Fast evaluation
            for p in range(len(candidates)):
                candidates[p].f = result[p]
                candidates[p].O[0] = result[p]
        else:
            # Full evaluation
            for p in range(len(candidates)):
                for io in range(self.objectives):
                    candidates[p].O[io] = result[p][io]
                candidates[p].f = np.dot(candidates[p].O, self.objective_weights)
                for ic in range(self.constraints):
                    candidates[p].C[ic] = result[p][self.objectives + ic]

        return candidates

    def collective_evaluation(self, candidates):

        if self.number_of_processes > 1:
            self._multiprocess_evaluate(candidates)
        else:
            if self.forward_unique_str:
                for p in range(len(candidates)):
                    candidates[p].evaluate(self._gen_unique_str())
            else:
                for p in range(len(candidates)):
                    candidates[p].evaluate()

    def multiprocess_evaluate_test(self, number_of_candidates, seed=None):
        # Function for testing purpose only - remove it!

        self._init_optimizer()
        if None:
            np.random.seed() # Possibe fix for some seed issues in parallel execution
        else:
            np.random.seed(seed)

        candidates = np.array([CandidateState(self) for c in range(number_of_candidates)], dtype=CandidateState)

        for p in range(number_of_candidates):
            candidates[p].X = np.random.uniform(self.lb, self.ub)

        self.collective_evaluation(candidates)

        return candidates



class CandidateState:
    """Candidate solution for optimization problem"""

    
    def __init__(self, optimizer: Optimizer):
        self._optimizer = optimizer
        self.X = np.zeros(optimizer.dimensions)
        self.O = np.zeros(optimizer.objectives) * np.nan
        self.C = np.zeros(optimizer.constraints) * np.nan
        self.f = np.nan
                    
        # Evaluation
        if self._optimizer.objectives == 1 and self._optimizer.constraints == 0:
            self._evaluate = self._eval_fast
        else:
            self._evaluate = self._eval_full
        
        # Comparison operators
        if self._optimizer.objectives == 1 and self._optimizer.constraints == 0:
            self._eq_fn = CandidateState._eq_fast
            self._lt_fn = CandidateState._lt_fast 
            #self.__gt__ = self._gt_fast
        else:
            self._eq_fn = CandidateState._eq_full
            self._lt_fn = CandidateState._lt_full 
            #self.__gt__ = self._gt_full
        
        
    def copy(self):
        
        cP = CandidateState(self._optimizer)        
        # Real copy for ndarrays
        cP.X = np.copy(self.X)
        cP.O = np.copy(self.O)
        cP.C = np.copy(self.C)
        cP.f = self.f
        return cP
    
    def __str__(self): 
        nch = np.max([np.size(self.X), np.size(self.O), np.size(self.C)]) 
        nch = np.min([8, nch])
        #print(nch)
        s = '-' * (12 * nch + nch + 20)
        
        for i in range(int(np.ceil(np.size(self.X)/nch))):
            if i == 0:
                s += '\n' + 'X: '.rjust(20)
            else:
                s += '\n' + ' ' * 20
            s += ' '.join([f'{x:12.5e}' for x in self.X[i*nch:(i+1)*nch]])
        
        """
        s += '\n' + 'Objectives: '.rjust(20) + ' '.join([f'{o:12.5e}' for o in self.O]) + \
             '\n' + 'Constraints: '.rjust(20) + ' '.join([f'{c:12.5e}' for c in self.C]) + \
        """
        alllbl = self._optimizer.objective_labels + self._optimizer.constraint_labels
        lblchr = np.max([len(lbl) for lbl in alllbl]) + 3
        for o, olbl in zip(self.O, self._optimizer.objective_labels):
            s += '\n' + f'{olbl}: '.rjust(lblchr) + f'{o:12.5e}' 
        for c, clbl in zip(self.C, self._optimizer.constraint_labels):
            s += '\n' + f'{clbl}: '.rjust(lblchr) + f'{c:12.5e}' 
        s += '\n' + 'Fitness: '.rjust(lblchr) + f'{self.f:12.5e}' + \
             '\n' + '-' * (12 * nch + nch + 20)
        return s
    
    # Equality operator
    def __eq__(self, other): 
        return self._eq_fn(self, other) 
    
    @staticmethod
    def _eq_fast(a, b): 
        return a.f == b.f

    @staticmethod
    def _eq_full(a, b):
        return np.sum(np.abs(a.X - b.X)) + np.sum(np.abs(a.O - b.O)) + np.sum(np.abs(a.C - b.C)) == 0.0

    # Inequality operator
    def __ne__(self, other):
        return self.f != other.f
        #return not self.__eq__(other)
    
    # Less-then operator
    def __lt__(self, other):
        return self._lt_fn(self, other)    

    def evaluate(self, s=None):
        return self._evaluate(s)

    @staticmethod
    def _lt_fast(a, b): 
        #print('_lt_fast')
        return a.f < b.f   
    
    @staticmethod     
    def _lt_full(a, b):  
        #print("_lt_full")               
        if np.sum(a.C > 0) == 0 and np.sum(b.C > 0) == 0: 
            # Both are feasible
            # Better candidate is the one with smaller fitness
            return a.f < b.f
        
        elif np.sum(a.C > 0) == np.sum(b.C > 0): 
            # Both are unfeasible and break same number of constraints
            # Better candidate is the one with smaller sum of constraint values
            return np.sum(a.C) < np.sum(b.C)
        
        else:
            # The number of unsatisfied constraints is not the same
            # Better candidate is the one which breaks fewer constraints
            return np.sum(a.C > 0) < np.sum(b.C > 0)
       
    
    def __gt__(self, other):
        #print('__gt__')
        #return self.f > other.f
        return not (self._lt_fn(self, other) or self._eq_fn(self, other))
    """
    def _gt_fast(self, other):
        return self.f > other.f
    def _gt_full(self, other): 
        return not (self.__eq__(other) or self.__lt__(other))
    """    
    
    def __le__(self, other):
        #print('__le__')
        #return self.f <= other.f
        return self._lt_fn(self, other) or self.__eq__(other)
    
    def __ge__(self, other):
        #print('__ge__')
        #return self.f >= other.f
        return self.__gt__(other) or self.__eq__(other)
    
    
    def set_X(self, X: np.ndarray):
        assert np.size(self.X) == np.size(X), 'Unexpected size of optimization vector X'
        self.X = X
        
    def _eval_full(self, s=None):
        
        #print('_eval_full')
        #print(self.X.shape)
        if s is None:
            oc = self._optimizer.evaluation_function(self.X)
        else:
            oc = self._optimizer.evaluation_function(self.X, s)
        #print(oc)

        for io in range(self._optimizer.objectives):
            self.O[io] = oc[io]
        
        self.f = np.dot(self.O, self._optimizer.objective_weights)
        
        for ic in range(self._optimizer.constraints):
            self.C[ic] = oc[self._optimizer.objectives + ic]

        #print('_eval_full', self.X[:5], self.f)
        return self.O, self.C, self.f
        
    def _eval_fast(self, s=None):

        if s is None:
            f = self._optimizer.evaluation_function(self.X)
        else:
            f = self._optimizer.evaluation_function(self.X, s)

        self.f = f
        self.O[0] = f 
        #print('_eval_fast', self.f, type(self.f))
        #input(' >> Press return to continue.')
        return self.O, self.C, self.f

class OptimizationRecord(CandidateState):
    
    def __init__(self, optimizer: Optimizer):
        super(OptimizationRecord, self).__init__(optimizer)        
        self.evaluation = None

class OptimizationResults:
    """Data holder for optimization results."""
    
    def __init__(self, optimizer: Optimizer):
        
        self.optimizer = optimizer
        self.cHistory = []
        
    def plot_convergence(self, axes=None):
        
        if axes is None:
            fig = plt.figure(constrained_layout=True)
            spec = gridspec.GridSpec(ncols=1, nrows=2, figure=fig)
            axo = fig.add_subplot(spec[0])
            axc = fig.add_subplot(spec[1], sharex=axo)
        else:
            axo, axc = axes
        
        for io in range(self.optimizer.objectives):
            O = np.array([cB.O[io] for cB in self.cHistory])
            axo.plot(np.arange(np.size(O)), O, label=self.optimizer.objective_labels[io], lw=1)
            
        F = np.array([cB.f for cB in self.cHistory])
        axo.plot(np.arange(np.size(F)), F, label='f', lw=2, ls='-', c='b')
         
        axo.legend()
        
        for ic in range(self.optimizer.constraints):
            C = np.array([cB.C[ic] for cB in self.cHistory])
            I = np.arange(np.size(C))
            axc.plot(I, C, label=self.optimizer.constraint_labels[io], lw=1)
            
            I = I[C > 0]
            C = C[C > 0]
            
            nc = np.sum(C>0)
            axc.plot(I[:nc], C[:nc], c='r', ls='-', lw=2)
            if nc == 1:
                axc.plot(I[:nc], C[:nc],'ro')
                
            
        axc.legend()
        
        axo.set_ylabel('Objectives')
        axc.set_xlabel('Iterations')
        axc.set_ylabel('Constraints')
        #ax.set_yscale('log')
        
        


# -*- coding: utf-8 -*-
#import indago.optimizer as optimizer
#import indago.pso as pso
#import indago.fwa as fwa
#import indago.abca as abca

#import optimizer
#import pso
#import fwa
#import abca
__version__ = '0.1.2'



# -*- coding: utf-8 -*-

import numpy as np
from .optimizer import Optimizer, OptimizationResults 


class LegacyPSO(Optimizer):
    """Particle Swarm Optimization class"""

    def __init__(self):
        """Initialization"""
        super(LegacyPSO, self).__init__()

        self.X = None
        self.X0 = None
        self.method = None
        self.iterations = None
        self.params = {}

    def _init_optimizer(self):

        super(LegacyPSO, self)._init_optimizer()
        
        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []

        if self.method == 'Vanilla':
            mandatory_params = 'inertia cognitive_rate social_rate'.split()
            optional_params = 'akb_fun_start akb_fun_stop'.split()
        elif self.method == 'TVAC':
            mandatory_params = 'inertia'.split()
            optional_params = 'akb_fun_start akb_fun_stop'.split()

        # print(defined_params, mandatory_params, optional_params)

        for param in mandatory_params:
            # if param not in defined_params:
            #    print('Error: Missing parameter (%s)' % param)
            assert param in defined_params, f'Error: Missing parameter {param}'

        for param in defined_params:
            if param not in mandatory_params and param not in optional_params:
                print(f'Warning: Excessive parameter {param}')

        """ Anakatabatic Inertia a.k.a. Polynomial PFIDI """
        if self.params['inertia'] == 'anakatabatic':
            assert 'akb_fun_start' in defined_params, \
            'Error: anakatabatic inertia requires akb_fun_start parameter'
            assert 'akb_fun_stop' in defined_params, \
            'Error: anakatabatic inertia requires akb_fun_stop parameter'

        # input(' >> Press return to continue')

        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)
        self.v_max = 0.2 * (self.ub - self.lb)

        self.X = np.empty([self.swarm_size, self.dimensions]) * np.nan
        self.V = np.empty([self.swarm_size, self.dimensions]) * np.nan
        self.F = np.empty([self.swarm_size]) * np.nan
        self.dF = np.empty([self.swarm_size]) * np.nan

        for p in range(self.swarm_size):
            self.X[p, :] = np.random.uniform(self.lb, self.ub)
            if not self.X0 is None:
                if p < np.shape(self.X0)[0]:
                    self.X[p] = self.X0[p]
            self.V[p, :] = np.random.uniform(-self.v_max, self.v_max)
            self.F[p] = self.objective(self.X[p, :])
            self.dF[p] = 0.0
        
        #print(self.X)
        #print(self.V)
        #input(' >> Press return to continue.')
        self.BX = np.copy(self.X)
        self.BF = np.copy(self.F)

        self.BI = np.zeros(self.swarm_size, dtype=int)
        self.TOPO = np.zeros([self.swarm_size, self.swarm_size], dtype=np.bool)

        self.topology(True)
        
        self.ET = np.zeros([self.iterations])
        

    def topology(self, reinitialize=True):
        # self.BI = np.argmin(self.BF)
        """ ring topology, neighborhood of size k """
        """
        k = 3
        for p in range(self.swarm_size):
            for p2 in range(p-k, p+1):
                if self.BF[p2] < self.BF[p]:
                    self.BI[p] = p2
        """
        k = 3
        if reinitialize:
            self.TOPO[:, :] = False
            for p in range(self.swarm_size):
                links = np.random.randint(self.swarm_size, size=k)
                self.TOPO[p, links] = True
                self.TOPO[p, p] = True
            # self.TOPO[:,:] = True # gbest

        for p in range(self.swarm_size):
            links = np.where(self.TOPO[p, :])[0]
            best = np.argmin(self.BF[links])
            best = links[best]
            self.BI[p] = best

    def run(self, seed=None):
        super(LegacyPSO, self).run(seed=seed)
        self._init_optimizer()

        if self.verbose:
            if self.params['inertia'] == 'anakatabatic':
                self.report['theta'] = np.empty([self.swarm_size,
                                                 self.iterations])

        if 'inertia' in self.params.keys():
            w = self.params['inertia']
        if 'cognitive_rate' in self.params.keys():
            c1 = self.params['cognitive_rate']
        if 'cognitive_rate' in self.params.keys():
            c2 = self.params['social_rate']

        for i in range(self.iterations):
            R1 = np.random.uniform(0, 1, [self.swarm_size, self.dimensions])
            R2 = np.random.uniform(0, 1, [self.swarm_size, self.dimensions])

            """ LDIW """
            if self.params['inertia'] == 'LDIW':
                w = 1.0 - (1.0 - 0.4) * i / self.iterations

            """ TVAC """
            if self.method == 'TVAC':
                c1 = 2.5 - (2.5 - 0.5) * i / self.iterations
                c2 = 0.5 + (2.5 - 0.5) * i / self.iterations

            """ Anakatabatic Inertia """
            if self.params['inertia'] == 'anakatabatic':

                theta = np.arctan2(self.dF, np.min(self.dF))
                theta[theta < 0] = theta[theta < 0] + 2 * np.pi  # 3rd quadrant
                # fix for atan2(0,0)=0
                theta0 = theta < 1e-300
                theta[theta0] = np.pi / 4 + \
                    np.random.rand(np.sum(theta0)) * np.pi
                w_start = self.params['akb_fun_start'](theta)
                w_stop = self.params['akb_fun_stop'](theta)
                w = w_start * (1 - i / self.iterations) \
                    + w_stop * (i / self.iterations)
                w = np.tile(np.array([w]).T, [1, self.dimensions])

                if self.verbose:
                    self.report['theta'][:, i] = theta

            self.tic()
            self.V = w * self.V + c1 * R1 * \
                (self.BX - self.X) + c2 * R2 * (self.BX[self.BI, :] - self.X)
            self.X = self.X + self.V
            #for p in range(self.swarm_size):
            #     self.V[p,:] = w * self.V[p,:] + c1 * R1[p,:] * \
            #     (self.BX[p,:] - self.X[p,:]) + c2 * R2[p,:] * (self.BX[self.BI[p], :] - self.X[p,:])
            #     self.X[p,:] = self.X[p,:] + self.V[p,:]
                
            
            for p in range(self.swarm_size):
                ilb = self.X[p, :] < self.lb
                self.X[p, ilb] = self.lb[ilb]
                iub = self.X[p, :] > self.ub
                self.X[p, iub] = self.ub[iub]

            for p in range(self.swarm_size):

                F = self.objective(self.X[p, :])
                self.dF[p] = self.F[p] - F
                self.F[p] = F

            for p in range(self.swarm_size):
                if self.F[p] <= self.BF[p]:
                    self.BF[p] = np.copy(self.F[p])
                    self.BX[p, :] = np.copy(self.X[p, :])
                
                
            best = np.argmin(self.BF)
            #self.results.X[i, :] = self.BX[best, :].copy()
            #self.results.F[i] = self.BF[best].copy()
            
            # Measure and record execution time
            self.ET[i] = self.toc('   PSO', silent=True)

            self.topology(np.min(self.dF) <= 0.0)

            # print(i, np.min(self.BF[self.BI]))

        # print(self.BX[self.BI])
        # print(self.BF[self.BI])
        # print(np.min(self.BF[self.BI]))

        best = np.argmin(self.BF)
        # print(self.BF)

        if self.verbose:
            return self.BF[best], self.BX[best], self.report
        else:
            # return np.min(self.BF[self.BI])
            return self.BF[best], self.BX[best]



# Source: https://github.com/rwuilbercq/Hive/blob/master/Hive/Hive.py

import random
import sys
import numpy as np
from copy import deepcopy
from .optimizer import Optimizer


class ABC(Optimizer):
    """Artificial Bee Colony Algorithm class"""

    def __init__(self):
        """Initialization"""
        super(ABC, self).__init__()

        self.dimensions = None
        self.population_size = None
        self.iterations = None
        self.max_trials = None

    class Bee():
        """Bee object"""

        def __init__(self, lb, ub, objective):
            self.vector = np.zeros(len(lb))

            self.objective = objective

            for i in range(len(lb)):
                self.vector[i] = lb[i] + random.random() * (ub[i] - lb[i])

            self.evaluate()

            # Initialises trial limit counter
            self.counter = 0

        def evaluate(self):
            self.value = self.objective(self.vector)

            if self.value >= 0:
                self.fitness = 1 / (1 + self.value)
            else:
                self.fitness = 1 + abs(self.value)

    def _init_optimizer(self):
        """
        Instantiates a bee hive object.

        1. INITIALISATION PHASE.
        -----------------------
        The initial population of bees should cover the entire search space as much as possible by randomizing individuals within the search space constrained by the prescribed lower and upper bounds.
        """

        # Calculate number of employees
        self.hive_size = int(self.population_size + self.population_size % 2)

        # initialises current best and its a solution vector
        self.best = sys.float_info.max
        self.solution = None

        # Create a bee hive
        self.population = [self.Bee(self.lb, self.ub, self.objective) for _ in range(self.hive_size)]

        # Find best bee candidate
        self.__find_best_bee()

        # Compute selection probability
        self.__compute_probability()

    def run(self, seed=None):
        super(LegacyPSO, self).run(seed=seed)
        self._init_optimizer()

        cost = {}
        cost["best"] = np.zeros(self.iterations)
        cost["mean"] = np.zeros(self.iterations)

        for i in range(self.iterations):
            # Employees phase
            for j in range(self.hive_size):
                self.__send_employee(j)

            # Onlookers phase
            self.__send_onlookers()

            # Scouts phase
            self.__send_scout()

            # Compute best solution
            self.__find_best_bee()

            # Convergence information
            cost["best"][i] = self.best
            cost["mean"][i] = sum([bee.value for bee in self.population]) / self.hive_size

        return self.best

    def __find_best_bee(self):
        vals = [bee.value for bee in self.population]
        ind = vals.index(min(vals))

        if vals[ind] < self.best:
            self.best = vals[ind]
            self.solution = self.population[ind].vector

    def __compute_probability(self):
        """
        Computes the relative chance that a given solution vector is chosen by an onlooker bee after the Waggle dance ceremony when employed bees are back within the hive.
        """

        # Rtrieves fitness of bees within the hive
        values = [bee.fitness for bee in self.population]
        max_value = max(values)

        # computes probalities the way Karaboga does in his classic ABC implementation
        self.probabilities = [0.9 * v / max_value + 0.1 for v in values]

        # Returns intervals of probabilities
        return [sum(self.probabilities[:i + 1]) for i in range(self.hive_size)]

    def __send_employee(self, ind):
        """
        2. SEND EMPLOYED BEES PHASE.
        ---------------------------
        During this 2nd phase, new candidate solutions are produced for each employed bee by cross-over and mutation of the employees. If the modified vector of the mutant bee solution is better than that of the original bee, the new vector is assigned to the bee.
        """

        # Copying current bee solution vector
        zombee = deepcopy(self.population[ind])

        # Draws a dimension to be crossed-over and mutated
        d = random.randint(0, self.dimensions - 1)

        # Select another bee
        bee_ind = ind
        while bee_ind == ind:
            bee_ind = random.randint(0, self.hive_size - 1)

        # Produces a mutant based on a current bee and bee's friend
        zombee.vector[d] = self.__mutate(d, ind, bee_ind)

        # Evaluate mutation
        zombee.evaluate()

        if zombee.fitness > self.population[ind].fitness:
            self.population[ind] = deepcopy(zombee)
            self.population[ind].counter = 0
        else:
            self.population[ind].counter += 1

    def __mutate(self, dim, current_bee, other_bee):
        """Mutates a given solution vector"""
        value = self.population[current_bee].vector[dim] + (random.random() - 0.5) * 2 * (self.population[current_bee].vector[dim] - self.population[other_bee].vector[dim])

        if value < self.lb[dim]:
            return self.lb[dim]

        if value > self.ub[dim]:
            return self.ub[dim]

        return value

    def __send_onlookers(self):
        """
        3. SEND ONLOOKERS PHASE.
        -----------------------
        We define as many onlooker bees as there are employed bees in the hive since onlooker bees will attempt to locally improve the solution path of the employed bee they have decided to follow after the waggle dance phase.
        If they improve it, they will communicate their findings to the bee they initially watched "waggle dancing".
        """

        num_onlookers = 0
        beta = 0
        while num_onlookers < self.hive_size:
            # Random number from U[0, 1]
            phi = random.random()

            # Increments roulette wheel parameter beta
            beta += phi * max(self.probabilities)
            beta %= max(self.probabilities)

            # Selects a new onlooker based on waggle dance
            index = self.__waggle_dance(beta)

            # Sends new onlooker
            self.__send_employee(index)

            # Increments number of onlookers
            num_onlookers += 1

    def __waggle_dance(self, beta):
        """
        4. WAGGLE DANCE PHASE.
        ---------------------
        During this 4th phase, onlooker bees are recruited using a roulette wheel selection.
        This phase represents the "waggle dance" of honey bees (i.e. figure- eight dance). By performing this dance, successful foragers (i.e. "employed" bees) can share, with other members of the colony, information about the direction and distance to patches of flowers yielding nectar and pollen, to water sources, or to new nest-site locations.
        During the recruitment, the bee colony is re-sampled in order to mostly keep, within the hive, the solution vector of employed bees that have a good fitness as well as a small number of bees with lower fitnesses to enforce diversity.
        """

        probs = self.__compute_probability()

        for i in range(self.hive_size):
            if beta < probs[i]:
                return i

    def __send_scout(self):
        """
        5. SEND SCOUT BEE PHASE.
        -----------------------
        Identifies bees whose abandonment counts exceed preset trials limit, abandons it and creates a new random bee to explore new random area of the domain space.
        In real life, after the depletion of a food nectar source, a bee moves on to other food sources.
        By this means, the employed bee which cannot improve their solution until the abandonment counter reaches the limit of trials becomes a scout bee. Therefore, scout bees in ABC algorithm prevent stagnation of employed bee population.
        Intuitively, this method provides an easy means to overcome any local optima within which a bee may have been trapped.
        """


        # Retrieves the number of trials for all bees
        trials = [self.population[i].counter for i in range(self.hive_size)]

        # Identifies the bee with the greatest number of trials
        ind = trials.index(max(trials))

        # Checks if its number of trials exceeds the pre-set maximum number of trials
        if trials[ind] > self.max_trials:
            # Creates a new scout bee randomly
            self.population[ind] = self.Bee(self.lb, self.ub, self.objective)

            # sends scout bee to exploit its solution vector
            self.__send_employee(ind)



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 21 15:57:13 2017

@author: stefan
"""

import numpy as np
import matplotlib.pyplot as plt
# import scipy.optimize as opt


class ShortestPath():
    """Shortest path test function"""

    def __init__(self, case):

        if case == 'case_1.1':
            self.A = [0, 0]
            self.B = [100, 100]
            self.XC = [50]
            self.YC = [55]
            self.R = [20]

        elif case == 'case_2.1':
            self.A = [0, 00]
            self.B = [150, 150]
            self.XC = [50, 100]
            self.YC = [50, 100]
            self.R = [15, 25]

        elif case == 'case_2.2':
            self.A = [0, 00]
            self.B = [150, 150]
            self.XC = [50, 100]
            self.YC = [100, 50]
            self.R = [30, 35]

        elif case == 'case_2.3':
            self.A = [0, 00]
            self.B = [150, 150]
            self.XC = [50, 100]
            self.YC = [100, 50]
            self.R = [40, 32]

        elif case == 'case_2.4':
            self.A = [0, 00]
            self.B = [150, 150]
            self.XC = [45, 100]
            self.YC = [100, 45]
            self.R = [40, 32]

        elif case == 'case_2.5':
            self.A = [0, 00]
            self.B = [150, 150]
            self.XC = [30, 120]
            self.YC = [70, 80]
            self.R = [44, 44]
            
        elif case == 'case_3.1':
            self.A = [20, 20]
            self.B = [150, 150]
            self.XC = [50, 110, 120]
            self.YC = [60, 75, 125]
            self.R = [15, 25, 10]

        elif case == 'case_3.2':
            self.A = [20, 20]
            self.B = [150, 150]
            self.XC = [50, 110, 120]
            self.YC = [60, 75, 125]
            self.R = [12, 18, 7]

        elif case == 'case_6.1':
            self.A = [20, 20]
            self.B = [150, 150]
            self.XC = [135, 78, 57, 102, 147, 103]
            self.YC = [143, 123, 55, 57, 65, 97]
            self.R = [8, 11, 9, 8, 11, 10]

        else:
            assert False, 'Unknown ShortestPath case (%s)' % case

        self.case = case

    def obj_cnstr(self, phi):
        
        x, y, L, D = self.generate_path(phi)

        return L, D
    
    def penalty(self, phi):

        x, y, L, D = self.generate_path(phi)

        if D > 0.0:
            P = 10 * D
        else:
            P = 0.0

        f = L + P

        return f

    def generate_path(self, phi):

        phi = np.deg2rad(phi)
        n = np.size(phi) + 1
        beta = np.zeros(n)
        for i in range(1, n):
            beta[i] = beta[i - 1] + phi[i - 1]

        x1, y1 = np.zeros(n + 1), np.zeros(n + 1)
        x1[0], y1[0] = self.A[0], self.A[1]
        for i in range(1, n + 1):
            x1[i] = x1[i - 1] + np.cos(beta[i - 1])
            y1[i] = y1[i - 1] + np.sin(beta[i - 1])

        alpha = np.arctan2(self.B[1] - self.A[1], self.B[0] - self.A[0])
        # print(np.rad2deg(alpha))

        alpha1 = np.arctan2(y1[-1] - self.A[1], x1[-1] - self.A[0])
        # print(np.rad2deg(alpha1))

        beta2 = np.zeros(n)
        beta2[0] = alpha - alpha1
        for i in range(1, n):
            beta2[i] = beta2[i - 1] + phi[i - 1]

        x2, y2 = np.zeros(n + 1), np.zeros(n + 1)
        x2[0], y2[0] = self.A[0], self.A[1]
        for i in range(1, n + 1):
            x2[i] = x2[i - 1] + np.cos(beta2[i - 1])
            y2[i] = y2[i - 1] + np.sin(beta2[i - 1])

        ll = np.sqrt((self.A[0] - self.B[0])**2 + (self.A[1] - self.B[1])**2)
        l2 = np.sqrt((self.A[0] - x2[-1])**2 + (self.A[1] - y2[-1])**2)
        k = ll / l2

        x, y = np.zeros(n + 1), np.zeros(n + 1)
        x[0], y[0] = self.A[0], self.A[1]
        for i in range(1, n + 1):
            x[i] = x[i - 1] + k * np.cos(beta2[i - 1])
            y[i] = y[i - 1] + k * np.sin(beta2[i - 1])

        # Calculating the length of path inside obstacles

        x1 = x[0:-1]
        y1 = y[0:-1]
        x2 = x[1:]
        y2 = y[1:]

        D = 0.0
        for xc, yc, r in zip(self.XC, self.YC, self.R):
            # print('Circle (%f, %f, %f)' % (xc, yc, r))
            a = (x2 - x1)**2 + (y2 - y1)**2
            b = 2 * (x2 - x1) * (x1 - xc) + 2 * (y2 - y1) * (y1 - yc)
            c = x1**2 + xc**2 - 2 * x1 * xc + y1**2 + yc**2 - 2 * y1 * yc - r**2

            d = b**2 - 4 * a * c
            a = a[d >= 0.0]
            b = b[d >= 0.0]
            d = d[d >= 0.0]

            t1 = (-b - np.sqrt(d)) / (2 * a)
            t2 = (-b + np.sqrt(d)) / (2 * a)

            # print(t1)
            # print(t2)

            t1[t1 < 0] = 0
            t2[t2 < 0] = 0
            t1[t1 > 1] = 1
            t2[t2 > 1] = 1

            # print(t1)
            # print(t2)

            # D += np.sum(np.abs(t2 - t1)) * k
            D += np.sum(t2 - t1) * k
            # print(t2 - t1, D)
            # print()

        L = k * n

        return x, y, L, D

    def draw_obstacles(self, ax):

        ax.plot(self.A[0], self.A[1], 'bo')
        ax.plot(self.B[0], self.B[1], 'go')

        for xc, yc, r in zip(self.XC, self.YC, self.R):
            circ = plt.Circle((xc, yc), r, color='grey', alpha=0.2)
            ax.add_artist(circ)

    def draw_path(self, phi, ax=None, label=None, c=None, ls='-', lw=1, a=1.0):

        x, y, L, D = self.generate_path(phi)

        if ax is None:
            plt.figure()
            ax = plt.gca()
            self.draw_obstacles(ax)
            ax.set_title('L: %.2f, D: %.2f' % (L, D))

        ax.plot(x, y, label=label, c=c, ls=ls, lw=lw, alpha=a)
        ax.axis('image')


if __name__ == '__main__':
    sp = ShortestPath('case_6.1')
    x = np.random.uniform(-30, 30, 30)
    # x = np.array([  6.72344496, 17.28081076, 2.69191604, 6.32926472, -8.14712714, 29.05435553, -1.57803371, -12.51337557, -2.24810493, 10.01112831])
    print(x)
    sp.draw_path(x)



# -*- coding: utf-8 -*-

import numpy as np
from .optimizer import Optimizer, CandidateState, OptimizationResults 
import copy
from scipy.interpolate import interp1d # need this for akb_model


class Particle(CandidateState):
    """PSO Particle class"""
    
    def __init__(self, optimizer: Optimizer):
        CandidateState.__init__(self, optimizer)
        #super(Particle, self).__init__(optimizer) # ugly version of the above
        
        self.V = np.zeros([optimizer.dimensions]) * np.nan

class PSO(Optimizer):
    """Particle Swarm Optimization class"""

    def __init__(self):
        """Initialization"""
        Optimizer.__init__(self)
        #super(PSO, self).__init__() # ugly version of the above

        self.X0 = None
        self.method = 'Vanilla'
        self.swarm_size = None
        self.params = {}

    def _check_params(self):
        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []

        if self.method == 'Vanilla':
            mandatory_params = 'inertia cognitive_rate social_rate'.split()
            optional_params = 'akb_model akb_fun_start akb_fun_stop'.split()
        elif self.method == 'TVAC':
            mandatory_params = 'inertia'.split()
            optional_params = 'akb_model akb_fun_start akb_fun_stop'.split()

        Optimizer._check_params(self, mandatory_params, optional_params, defined_params)


        """ Anakatabatic Inertia a.k.a. Polynomial PFIDI """
        if self.params['inertia'] == 'anakatabatic':
            assert ('akb_fun_start' in defined_params \
                    and 'akb_fun_stop' in defined_params) \
                    or 'akb_model' in defined_params, \
                    'Error: Anakatabatic inertia requires either akb_model parameter or akb_fun_start and akb_fun_stop parameters'
                    
            if 'akb_model' in defined_params:

                if self.params['akb_model'] in ['FlyingStork', 'MessyTie', 'RightwardPeaks', 'OrigamiSnake']:   # w-list-based named akb_models                
                    if self.params['akb_model'] == 'FlyingStork':
                        w_start = [-0.86, 0.24, -1.10, 0.75, 0.72]
                        w_stop = [-0.81, -0.35, -0.26, 0.64, 0.60]
                        splinetype = 'linear'
                        if self.method != 'Vanilla':
                            print('Warning: akb_model \'FlyingStork\' was designed for Vanilla PSO')                    
                    elif self.params['akb_model'] == 'MessyTie':
                        w_start = [-0.62, 0.18, 0.65, 0.32, 0.77]
                        w_stop = [0.36, 0.73, -0.62, 0.40, 1.09]
                        splinetype = 'linear'
                        if self.method != 'Vanilla':
                            print('Warning: akb_model \'MessyTie\' was designed for Vanilla PSO')   
                    elif self.params['akb_model'] == 'RightwardPeaks':
                        w_start = [-1.79, -0.33, 2.00, -0.67, 1.30]
                        w_stop = [-0.91, -0.88, -0.84, 0.67, -0.36]
                        splinetype = 'linear'
                        if self.method != 'TVAC':
                            print('Warning: akb_model \'RightwardPeaks\' was designed for TVAC PSO')
                    elif self.params['akb_model'] == 'OrigamiSnake':
                        w_start = [-1.36, 2.00, 1.00, -0.60, 1.22]
                        w_stop = [0.30, 1.03, -0.21, 0.40, 0.06]
                        splinetype = 'linear'
                        if self.method != 'TVAC':
                            print('Warning: akb_model \'OrigamiSnake\' was designed for TVAC PSO')                      # code shared for all w-list-based named akb_models
                    Th = np.linspace(np.pi/4, 5*np.pi/4, 5)
                    self.params['akb_fun_start'] = \
                                        interp1d(Th, w_start, kind=splinetype)
                    self.params['akb_fun_stop'] = \
                                        interp1d(Th, w_stop, kind=splinetype) 
                else:
                    if self.params['akb_model'] != 'Languid':
                        print('Warning: Unknown akb_model. Defaulting to \'Languid\'')
                        self.params['akb_model'] = 'Languid'
                if self.params['akb_model'] == 'Languid':
                    def akb_fun_languid(Th):
                        w = (0.72 + 0.05) * np.ones_like(Th)
                        for i, th in enumerate(Th):
                            if th < 4*np.pi/4: 
                                w[i] = 0
                        return w
                    self.params['akb_fun_start'] = akb_fun_languid
                    self.params['akb_fun_stop'] = akb_fun_languid 
                
        
    def _init_method(self):
        self._init_optimizer()
        
        # Bounds for position and velocity
        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)
        self.v_max = 0.2 * (self.ub - self.lb)

        # Generate a swarm
        self.cS = np.array([Particle(self) for c in range(self.swarm_size)], dtype=Particle)
        
        # Prepare arrays
        self.dF = np.empty([self.swarm_size]) * np.nan

        # Generate initial positions
        for p in range(self.swarm_size):
            
            # Random position
            self.cS[p].X = np.random.uniform(self.lb, self.ub)
            
            # Using specified particles initial positions
            if self.X0 is not None:
                if p < np.shape(self.X0)[0]:
                    self.cS[p].X = self.X0[p]
                    
            # Generate velocity
            #self.V[p, :] = np.random.uniform(-self.v_max, self.v_max)
            self.cS[p].V = np.random.uniform(-self.v_max, self.v_max)

            # No fitness change at the start
            self.dF[p] = 0.0

        # Evaluate
        self.collective_evaluation(self.cS)
        
        # Use initial particles as best ones
        self.cB = np.array([cP.copy() for cP in self.cS], dtype=CandidateState)
        
        # Update the overall best
        #self.p_best = np.argmin([cP.f for cP in self.cB])
        self.p_best = np.argmin(self.cB)
            
        # Update history
        self.results = OptimizationResults(self)
        self.results.cHistory = [self.cB[self.p_best].copy()]
        
        self.BI = np.zeros(self.swarm_size, dtype=int)
        self.TOPO = np.zeros([self.swarm_size, self.swarm_size], dtype=np.bool)

        self.reinitialize_topology()
        self.find_neighbourhood_best()
        
        self.ET = np.zeros([self.iterations])
        
    def reinitialize_topology(self, k=3):
        self.TOPO[:, :] = False
        for p in range(self.swarm_size):
            links = np.random.randint(self.swarm_size, size=k)
            self.TOPO[p, links] = True
            self.TOPO[p, p] = True

    def find_neighbourhood_best(self):
        for p in range(self.swarm_size):
            links = np.where(self.TOPO[p, :])[0]
            #best = np.argmin(self.BF[links])
            p_best = np.argmin(self.cB[links])
            p_best = links[p_best]
            self.BI[p] = p_best

    def run(self, seed=None):
        Optimizer.run(self, seed=seed)
        self._check_params()
        self._init_method()

        if self.verbose:
            if self.params['inertia'] == 'anakatabatic':
                self.report['theta'] = np.empty([self.swarm_size,
                                                 self.iterations])

        if 'inertia' in self.params.keys():
            w = self.params['inertia']
        if 'cognitive_rate' in self.params.keys():
            c1 = self.params['cognitive_rate']
        if 'cognitive_rate' in self.params.keys():
            c2 = self.params['social_rate']

        for i in range(self.iterations):
            R1 = np.random.uniform(0, 1, [self.swarm_size, self.dimensions])
            R2 = np.random.uniform(0, 1, [self.swarm_size, self.dimensions])

            """ LDIW """
            if self.params['inertia'] == 'LDIW':
                w = 1.0 - (1.0 - 0.4) * i / self.iterations

            """ TVAC """
            if self.method == 'TVAC':
                c1 = 2.5 - (2.5 - 0.5) * i / self.iterations
                c2 = 0.5 + (2.5 - 0.5) * i / self.iterations

            """ Anakatabatic Inertia """
            if self.params['inertia'] == 'anakatabatic':

                theta = np.arctan2(self.dF, np.min(self.dF))
                theta[theta < 0] = theta[theta < 0] + 2 * np.pi  # 3rd quadrant
                # fix for atan2(0,0)=0
                theta0 = theta < 1e-300
                theta[theta0] = np.pi / 4 + \
                    np.random.rand(np.sum(theta0)) * np.pi
                w_start = self.params['akb_fun_start'](theta)
                w_stop = self.params['akb_fun_stop'](theta)
                #print(w_start)
                w = w_start * (1 - i / self.iterations) \
                    + w_stop * (i / self.iterations)

                if self.verbose:
                    self.report['theta'][:, i] = theta
            
            w = w * np.ones(self.swarm_size) # ensure w is a vector
            
            # Calculate new velocity and new position
            
            for p, cP in enumerate(self.cS):

                self.cS[p].V = w[p] * self.cS[p].V + \
                               c1 * R1[p, :] * (self.cB[p].X - self.cS[p].X) + \
                               c2 * R2[p, :] * (self.cB[self.BI[p]].X - self.cS[p].X)
                        
                self.cS[p].X = self.cS[p].X + self.cS[p].V  
                
                # Correct position to the bounds
                cP.X = np.clip(cP.X, self.lb, self.ub)
#                ilb = cP.X < self.lb
#                cP.X[ilb] = self.lb[ilb]
#                iub = cP.X > self.ub
#                cP.X[iub] = self.ub[iub]         

                
            # Get old fitness
            f_old = np.array([cP.f for cP in self.cS])

            # Evaluate swarm
            self.collective_evaluation(self.cS)

            for p, cP in enumerate(self.cS):
                # Calculate dF
                self.dF[p] = cP.f - f_old[p]
                
            for p, cP in enumerate(self.cS):
                # Update personal best
                if cP <= self.cB[p]:
                    self.cB[p] = cP.copy()  
            
            # Update the overall best
            self.p_best = np.argmin(self.cB)
                        
            # Update history
            self.results.cHistory.append(self.cB[self.p_best].copy())
            if self.verbose:
                print(i, self.cB[self.p_best].f)
            
            # Stop if fitness threshold achieved
            if self.cB[self.p_best].f < self.target_fitness:
                break
            
            # Update swarm topology
            if np.max(self.dF) >= 0.0:
                self.reinitialize_topology()
                
            # Find best particles in neighbourhood 
            self.find_neighbourhood_best()

        return self.cB[self.p_best]
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 10 14:42:11 2019

@author: stefan
"""

import setuptools
from distutils.core import setup

with open("readme.md", "r") as fh:
    long_description = fh.read()
    
setup(name='Indago',
      version='0.1.4',
      description='Numerical optimization framework',
      author='sim.riteh.hr',
      author_email='stefan.ivic@riteh.hr',
      url='http://sim.riteh.hr/',
      py_modules=['indago',
                  'indago.optimizer',
                  'indago.pso',
                  'indago.fwa',
                  'indago.abca',
                  'indago.ssa',
                  'indago.direct_search',
                  'indago.benchmarks'],
      setup_requires=['wheel'],
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=setuptools.find_packages(),
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          ],
      python_requires='>=3.6',
)



# -*- coding: utf-8 -*-

import numpy as np
from .optimizer import Optimizer, CandidateState, OptimizationResults 
from scipy.special import gamma

"""
Jain, M., Singh, V., & Rani, A. (2019). A novel nature-inspired algorithm for 
optimization: Squirrel search algorithm. Swarm and evolutionary computation, 
44, 148-175.
"""


class FlyingSquirrel(CandidateState):
    """SSA agent class"""
    
    def __init__(self, optimizer: Optimizer):
        CandidateState.__init__(self, optimizer)


class SSA(Optimizer):
    """Squirrel Search Algorithm class"""

    def __init__(self):
        """Initialization"""
        
        Optimizer.__init__(self)

        self.X = None
        self.X0 = None
        self.method = 'Vanilla'
        self.swarm_size = None
        self.params = {}
        
        # Default params
        self.Pdp = 0.1 # predator_presence_probability
        self.Gc = 1.9 # gliding_constant
        self.dg_lim = [0.5, 1.11] # gliding_distance_limits (for random dg)

    def _check_params(self):

        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []

        if self.method == 'Vanilla':
            mandatory_params = 'acorn_tree_attraction'.split()
            optional_params = 'predator_presence_probability gliding_constant \
                                gliding_distance_limits'.split()

        Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
        
    def _init_method(self):
        
        self._init_optimizer()
        
        # Bounds for position and velocity
        self.lb = np.array(self.lb)
        self.ub = np.array(self.ub)

        # Generate a swarm of FS
        self.cS = np.array([FlyingSquirrel(self) for c in range(self.swarm_size)], \
                            dtype=FlyingSquirrel)
        
        # Generate initial positions
        for p in range(self.swarm_size):            
            # Random position
            self.cS[p].X = np.random.uniform(self.lb, self.ub)           
            # Using specified particles initial positions
            if self.X0 is not None:
                if p < np.shape(self.X0)[0]:
                    self.cS[p].X = self.X0[p]                                          
            # Evaluate one by one
            #self.cS[p].evaluate()   
        # Evaluate
        self.collective_evaluation(self.cS)
        
        # Find the overall best, i.e. FSht (hickory nut tree)
        self.best = np.sort(self.cS)[0]
            
        # Update history
        self.results = OptimizationResults(self)
        self.results.cHistory = [self.best]
        
    def run(self, seed=None):
        
        Optimizer.run(self, seed=seed)
        self._check_params()
        self._init_method()
      
        # Load params
        if 'acorn_tree_attraction' in self.params.keys():
            # part of FSnt moving to FSat
            # ATA=0 (all move to FSht) - emphasize local search
            # ATA=1 (all move to FSat's) - emphasize global search
            self.ATA = self.params['acorn_tree_attraction']
        if 'predator_presence_probability' in self.params.keys():
            self.Pdp = self.params['predator_presence_probability']
        if 'gliding_constant' in self.params.keys():
            self.Gc = self.params['gliding_constant']
        if 'gliding_distance_limits' in self.params.keys():
            self.dg_lim = self.params['gliding_distance_limits']
            
        def Levy():
            ra, rb = np.random.normal(0, 1), np.random.normal(0, 1)
            beta = 1.5
            sigma = ((gamma(1 + beta) * np.sin(np.pi * beta / 2)) / \
                     gamma((1 + beta) / 2) * beta * 2**((beta - 1)/2)) **(1 / beta)
            return 0.01 * (ra * sigma) / (np.abs(rb)**(1 / beta))

        for i in range(self.iterations):
            
            # Categorizing FS's
            FSht = np.sort(self.cS)[0] # best FS (hickory nut trees)
            FSat = np.sort(self.cS)[1:4] # good FS (acorn trees)
            FSnt = np.sort(self.cS)[5:] # bad FS (normal trees)
            
            """
            # Moving FSnt - cascading strategy
            # move principally to FSat; 
            # with probability = (1-Pdp)*Pdp = 0.09 move to Fsht
            for fs in FSnt:
                if np.random.rand() >= self.Pdp: # move towards FSat
                    dg = np.random.uniform(self.dg_lim[0], self.dg_lim[1])
                    fs.X = fs.X + dg * self.Gc * \
                            (np.random.choice(FSat).X - fs.X)
                elif np.random.rand() >= self.Pdp: # move towards FSht
                    dg = np.random.uniform(self.dg_lim[0], self.dg_lim[1])
                    fs.X = fs.X + dg * self.Gc * (FSht.X - fs.X)
                else: # not moving, i.e. respawning randomly
                    fs.X = np.random.uniform(self.lb, self.ub)
            """
            
            # Moving FSnt
            Nnt2at = int(np.size(FSnt) * self.ATA) # attracted to acorn trees
            np.random.shuffle(FSnt)
            for fs in FSnt[:Nnt2at]:
                if np.random.rand() >= self.Pdp: # move towards FSat
                    dg = np.random.uniform(self.dg_lim[0], self.dg_lim[1])
                    fs.X = fs.X + dg * self.Gc * \
                            (np.random.choice(FSat).X - fs.X)
                else: # not moving, i.e. respawning randomly
                    fs.X = np.random.uniform(self.lb, self.ub)
            for fs in FSnt[Nnt2at:]:
                if np.random.rand() >= self.Pdp: # move towards FSht
                    dg = np.random.uniform(self.dg_lim[0], self.dg_lim[1])
                    fs.X = fs.X + dg * self.Gc * (FSht.X - fs.X)
                else: # not moving, i.e. respawning randomly
                    fs.X = np.random.uniform(self.lb, self.ub)
            
            # Moving FSat
            for fs in FSat:
                if np.random.rand() >= self.Pdp: # move towards FSht
                    dg = np.random.uniform(self.dg_lim[0], self.dg_lim[1])
                    fs.X = fs.X + dg * self.Gc * (FSht.X - fs.X)
                else: # not moving, i.e. respawning randomly
                    fs.X = np.random.uniform(self.lb, self.ub)
            
            # Seasonal constants (for FSat)
            Sc = np.empty(3)
            for i, fs in enumerate(FSat):
                Sc[i] = np.sqrt(np.sum((fs.X - FSht.X)**2))
                
            # Minimum value of seasonal constant
            Scmin = 1e-6 / (365**(i / (self.iterations / 2.5))) # this is some black magic shit
            
            # Random-Levy relocation at the end of winter season
            if (Sc < Scmin).all():
                for fs in FSnt:
                    fs.X = self.lb + Levy() * (self.ub - self.lb)
            
            for cP in self.cS:
                # Correct position to the bounds
                cP.X = np.clip(cP.X, self.lb, self.ub)       

            # # Evaluate swarm, one by one
            # for cP in self.cS:
            #     cP.evaluate()
                
            # Evaluate swarm
            self.collective_evaluation(self.cS)
            
            # Update the overall best
            self.best = np.min(self.cS)
             
            # Update history
            self.results.cHistory.append(self.best.copy())

            if self.verbose:
                print(i, self.best.f)
                
            # Stop if fitness threshold achieved
            if self.best.f < self.target_fitness:
                break

        return self.best


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 10 13:43:18 2018

@author: Stefan
"""

# -*- coding: utf-8 -*-

import numpy as np
from .optimizer import Optimizer, CandidateState, OptimizationResults
import random


class FWA(Optimizer):
    """Firework Algorithm class"""

    def __init__(self):
        """Initialization"""
        Optimizer.__init__(self)

        self.X = None
        self.X0 = None
        self.cX = None
        self.method = 'Vanilla'
        self.params = {}

    def _check_params(self):

        defined_params = list(self.params.keys())
        mandatory_params, optional_params = [], []

        if self.method == 'Vanilla':
            mandatory_params = 'n m1 m2'.split()
            optional_params = ''.split()
        elif self.method == 'Rank':
            mandatory_params = 'n m1 m2'.split()
            optional_params = ''.split()
        else:
            assert False, f'Uknonwn method! {self.method}'

        Optimizer._check_params(self, mandatory_params, optional_params, defined_params)

    def _init_method(self):
        self._init_optimizer()

        self.cX = np.array([CandidateState(self) for p in range(self.params['n'])], dtype=CandidateState)
        
        # Generate initial positions
        for p in range(self.params['n']):
            
            # Random position
            self.cX[p].X = np.random.uniform(self.lb, self.ub, self.dimensions)
            
            # Using specified initial positions
            if self.X0 is not None:
                if p < np.shape(self.X0)[0]:
                    self.cX[p].X = self.X0[p]

        # Evaluate all
        self.collective_evaluation(self.cX)

        # Sort
        self.cX = np.sort(self.cX)
        
        # Initialize the results
        self.results = OptimizationResults(self)
        self.results.cHistory = [np.min(self.cX).copy()]


    def run(self, seed=None):
        Optimizer.run(self, seed=seed)
        self._check_params()
        self._init_method()
        
        n = self.params['n']

        for i in range(self.iterations):

            explosion_sparks = self.explosion()
            mutation_sparks = self.gaussian_mutation()

            #self.__mapping_rule(sparks, self.lb, self.ub, self.dimensions)
            for cS in (explosion_sparks + mutation_sparks):
                
                ilb = cS.X < self.lb
                cS.X[ilb] = self.lb[ilb]
                iub = cS.X > self.ub
                cS.X[iub] = self.ub[iub]

                self.cX = np.append(self.cX, [cS])

            self.collective_evaluation(self.cX)

            self.cX = np.sort(self.cX)[:n]

            # Update history
            self.results.cHistory.append(np.min(self.cX).copy())

            # Stop if fitness threshold achieved
            if np.min(self.cX).f < self.target_fitness:
                break
        
        return np.min(self.cX)
        # self.best = Gbest
        # self._set_Gbest(Gbest)

    def explosion(self):
        eps=0.001
        amp=10
        a=0.01
        b=10
        F = np.array([cP.f for cP in self.cX])
        fmin = np.min(F)
        fmax = np.max(F)
        
        explosion_sparks = []
        for p in range(self.params['n']):
               
            cFw = self.cX[p].copy()
            #print(cFw.X)
            
            if self.method == 'Vanilla':
                # Number of sparks
                n1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
                n1 = self.min_max_round(n1, self.params['m1'] * a, self.params['m2'] * b)
                
                # Amplitude
                A = amp * (cFw.f - fmin + eps) /  (np.sum(F - fmin) + eps)

                for j in range(n1):
                    for k in range(self.dimensions):
                        if (random.choice([True, False])):
                            cFw.X[k] += random.uniform(-A, A)
                    explosion_sparks.append(cFw.copy())
                
            if self.method == 'Rank':
                
                # Number of sparks
                #vn1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
                #vn1 = self.min_max_round(vn1, self.params['m1'] * a, self.params['m2'] * b)
                
                n1 = self.params['m1'] * (self.params['n'] - p)**1 / np.sum(np.arange(self.params['n']+1)**1)
                n1 = random.choice([int(np.floor(n1)), int(np.ceil(n1))])
                #print(self.cX[p].f, vn1, n1)
                
                # Amplitude
                #Ac = amp * (cFw.f - fmin + eps) / (np.sum(F - fmin) + eps)
                    
                #print('n1:', n1, 'A:', A)
                XX = np.array([cP.X for cP in self.cX])
                #print(XX.shape)
                
                # Uniform
                dev = np.std(XX, 0)
                avg_scale = np.average(np.sqrt(np.arange(self.params['n']) + 1))
                scale = np.sqrt(p + 1) / avg_scale
                
                #avg_scale = np.average(np.arange(self.params['n']) + 1)
                #scale = (p + 1) / avg_scale
                
                
                A = np.sqrt(12) / 2 * dev * scale
                A *= 1.5

                
                #cS = cFw.copy()
                for j in range(n1):
                    cFw.X = cFw.X + np.random.uniform(-A, A) * np.random.randint(0, 1, A.size)
                    
                    for k in range(self.dimensions):
                        if (random.choice([True, False])):
                            # Uniform
                            cFw.X[k] += np.random.uniform(-A[k], A[k])
                            # Normal
                            # cFw.X[k] += random.normal(-A[k], A[k])
                    
                    #print(cS.X)
                    explosion_sparks.append(cFw.copy())  

        #print('expl sparks:', len(explosion_sparks))
        #input(' > Press return to continue.')
        return explosion_sparks
    
    """
    def __explosion_operator(self, sparks, fw, function,
                             dimension, m, eps, amp, Ymin, Ymax, a, b):
        
        sparks_num = self.__round(m * (Ymax - function(fw) + eps) /
                                  (sum([Ymax - function(fwk) for fwk in self.X]) + eps), m, a, b)
        print(sparks_num)

        amplitude = amp * (function(fw) - Ymax + eps) / \
            (sum([function(fwk) - Ymax for fwk in self.X]) + eps)

        for j in range(int(sparks_num)):
            sparks.append(np.array(fw))
            for k in range(dimension):
                if (random.choice([True, False])):
                    sparks[-1][k] += random.uniform(-amplitude, amplitude)
    """
    
    def gaussian_mutation(self):
        
        mutation_sparks = []
        for j in range(self.params['m2']):
            cFw = self.cX[np.random.randint(self.params['n'])].copy()
            g = np.random.normal(1, 1)
            for k in range(self.dimensions):
                if(random.choice([True, False])):
                    cFw.X[k] *= g
            mutation_sparks.append(cFw)

        #print('mut sparks:', np.sort([p.f for p in mutation_sparks]))
        #print('mut sparks:', len(mutation_sparks))
        return mutation_sparks
            
    def __mapping_rule(self, sparks, lb, ub, dimension):
        for i in range(len(sparks)):
            for j in range(dimension):
                if(sparks[i].X[j] > ub[j] or sparks[i].X[j] < lb[j]):
                    sparks[i].X[j] = lb[j] + \
                        (sparks[i].X[j] - lb[j]) % (ub[j] - lb[j])

    def __selection(self, sparks, n, function):        

        self.collective_evaluation(sparks)
        self.cX = np.append(self.cX, sparks)


    def min_max_round(self, s, smin, smax):
        return int(np.round(np.min([np.max([s, smin]), smax])))

    def __round(self, s, m, a, b):
        if (s < a * m):
            return round(a * m)
        elif (s > b * m):
            return round(b * m)
        else:
            return round(s)
